"Reference module for interfacing with fbink. Cannot guarantee functionality is completely as advertised on device."

from PIL import Image
from typing import *

class ffi():
    def __init__(self):
        self.is_centered = False
        self.is_halfway = False

    @staticmethod
    def new(arg1):
        return lib

    @staticmethod
    def string(ctype: bytes) -> str:
        pass

    @staticmethod
    def list_types() -> tuple[
                    list[Literal["typedef_names"]],
                    list[Literal["names_of_structs"]],
                    list[Literal["names_of_unions"]]]:
        return

class lib():
    ##Gathered functions by calling help(lib) -> list is not exhaustive
    def __init__(self):
        pass

    @staticmethod
    def new(arg1):
        return ffi

    @staticmethod
    def fbink_init(arg1: int, fbinkconfig):
        """
        Initialize internal variables keeping track of the framebuffer's configuration and state, as well as the device's hardware.
        MUST be called at least *once* before any fbink_print*, fbink_dump/restore, fbink_cls or fbink_grid* functions.
        CAN safely be called multiple times,
            but doing so is only necessary if the framebuffer's state has changed (although fbink_reinit is preferred in this case),
            or if you modified one of the FBInkConfig fields that affects its results (listed below).
        Returns -(ENOSYS) if the device is unsupported (NOTE: Only on reMarkable!)
        fbfd:		Open file descriptor to the framebuffer character device,
        			if set to FBFD_AUTO, the fb is opened for the duration of this call.
        fbink_cfg:		Pointer to an FBInkConfig struct.
        			If you wish to customize them, the fields:
        			is_centered, fontmult, fontname, fg_color, bg_color,
        			no_viewport, is_verbose, is_quiet & to_syslog
        			MUST be set beforehand.
        			This means you MUST call fbink_init() again when you update them, too!
        			(This also means the effects from those fields "stick" across the lifetime of your application,
        			or until a subsequent fbink_init() (or effective fbink_reinit()) call gets fed different values).
        			NOTE: For fg_color & bg_color, see fbink_update_pen_colors().
        			NOTE: For is_verbose, is_quiet & to_syslog, see fbink_update_verbosity().
        NOTE: By virtue of, well, setting global variables, do NOT consider this thread-safe.
                The rest of the API should be, though, so make sure you init in your main thread *before* threading begins...
        NOTE: If you just need to make sure the framebuffer state is still up to date before an fbink_* call,
              (e.g., because you're running on a Kobo, which may switch from 16bpp to 32bpp, or simply change orientation),
               prefer using fbink_reinit instead of calling fbink_init *again*, as it's tailored for this use case.
               c.f., KFMon for an example of this use case in the wild.
        NOTE: You can perfectly well keep a few different FBInkConfig structs around, instead of modifying the same one over and over.
               Just remember that some fields *require* an fbink_init() call to be taken into account (see above),
               but if the only fields that differ don't fall into that category, you do *NOT* need an fbink_init() per FBInkConfig...
        """
        pass

    @staticmethod
    def fbink_version() -> bytes:
        pass

    @staticmethod
    def fbink_open():
        return True

    @staticmethod
    def fbink_close(arg1):
        pass

    @staticmethod
    def fbink_get_state(fbinkconfig, fbinkstate):
        pass

    @staticmethod
    def fbink_cls(arg1: int, fbinkconfig, fbinkrect, arg4: bool):
        pass

    @staticmethod
    def fbink_dump(arg1: int, fbinkdump):
        return
    
    @staticmethod
    def fbink_state_dump(fbinkconfig):
        "Dumps the fbink state in the console"
        return

    @staticmethod
    def fbink_free_dump_data(fbinkdump):
        return
    
    @staticmethod
    def fbink_get_fbinfo(fb_var_screeninfo, fb_fix_screeninfo):
        return
    
    @staticmethod
    def fbink_print_image(arg1, image_path, arg3, arg4, arg5):
        img = Image.open(image_path)
        img.show()

    @staticmethod
    def fbink_print_raw_data(arg1: int, raw_data, x: int, y: int, size_t, short1, short2, fbinkconfig):
        return
    
    @staticmethod
    def fbink_refresh(arg1: int, arg2: int, arg3: int, arg4: int, arg5: int, fbinkconfig):
        return
    
    @property
    def DATA(self):
        ##Doesn't work like this, each entry under DATA is it's own attribute
        return


    




"Functions that require pssm (i.e. a running inkBoard instance) to interface with the designer"

import logging
from typing import TYPE_CHECKING, Optional
import tkinter as tk
import tkthread


from PIL import ImageTk

import ttkbootstrap as ttk
from ttkbootstrap.tooltip import ToolTip

import inkBoard
from inkBoard import core as CORE

from PythonScreenStackManager import elements

from .. import const
from ..settings import EM_SETTINGS

from ..tkinter import window, functions as tk_functions
from ..tkinter.windows import TreeFrame, ElementWindow
from ..tkinter.widgets import Treeview

if TYPE_CHECKING:
    from PythonScreenStackManager.pssm.screen import PSSMScreen

_LOGGER = inkBoard.getLogger(__name__)


##Just call unbind first them
##Things to (re)bind: the listvariable/combobox
##Refresh Tree button
##Highlight Switch
##Backlight function
##Screenshot function

current_tree = None

MDI_TREE_ICONS = {}
ELEMENT_ICONS_TK = {}
ENTITY_ICONS_TK = {}

_ELEMENT_DICT = {}
_INDICATOR_RECTANGLES = []

ui_frame: ttk.Frame = window.children[const.UI_FRAME_NAME]
tree_frame: TreeFrame = ui_frame.children[const.TREE_FRAME_NAME]
canvas: tk.Canvas = window.children[const.CANVAS_NAME]

tree_frame.last_hover = False

##These do not get imported in the window thread.
def build_tk_icon(icon: str) -> ImageTk.PhotoImage:
    "Builds icon for the treeview in the correct thread"
    
    return tk_functions.build_tree_icon(icon)

ELEMENT_ICONS_TK["default"] = build_tk_icon(const.DEFAULT_ELEMENT_ICON)
ENTITY_ICONS_TK["default"] = build_tk_icon("mdi:one-up")

def add_element_icon(element : "elements.Element"):
    
    icon = element._emulator_icon
    icon = build_tk_icon(icon)
    ELEMENT_ICONS_TK[str(element.__class__.__name__)] = icon

@tkthread.called_on_main
def build_element_tree(screen: "PSSMScreen", open_items: bool = False):

    eltStack = screen.stack

    treeview = tree_frame.get_tree("Elements")

    open_init = open_items

    def make_layout_tree(layoutElt : "elements.Layout", parentiid : str):
        
        if isinstance(layoutElt, elements.TabPages):
            eltList = set(layoutElt.create_element_list()) | set(layoutElt.pageElements)
        else:
            eltList = layoutElt.create_element_list()


        for elt in eltList:
            elt : elements.Layout
            if not getattr(elt, "_isSubLayout", False) or (elt.__class__ not in  {elements.Layout, elements.baseelements._TileBase}):
                if elt.id == elt.unique_id:
                    eltname = elt.id.replace('_',' ')
                else:
                    eltname = elt.id
                elttype = str(elt.__class__.__name__)
                if elttype not in ELEMENT_ICONS_TK:
                    add_element_icon(elt)
                    _LOGGER.verbose(f"Made new icon for {elt}")
                icon = ELEMENT_ICONS_TK.get(elttype, ELEMENT_ICONS_TK["default"])
                iid = elt.id
                entity = getattr(elt,"entity","None")
                if not treeview.exists(iid):
                    treeview.insert(
                            parentiid,
                            tk.END,
                            iid = iid,
                            text=eltname,
                            values=(entity),
                            image=icon,
                            open=open_init
                        )
                    _ELEMENT_DICT[iid] = elt
                else:
                    if iid in treeview.get_children(parentiid):
                        pass
                    else:
                        treeview.reattach(iid, parentiid, tk.END)
            else:
                iid = parentiid

            if isinstance(elt,elements.Layout):
                make_layout_tree(elt, iid)

    for elt in eltStack:
        eltname = elt.id.replace('_',' ')
        elttype = str(elt.__class__.__name__)
        if elttype not in ELEMENT_ICONS_TK:
                add_element_icon(elt)
        icon = ELEMENT_ICONS_TK.get(elttype, ELEMENT_ICONS_TK["default"])
        iid = elt.id
        entity = getattr(elt,"entity","None")

        if not treeview.exists(iid):
            treeview.insert(
                "",
                tk.END,
                iid = iid,
                text=eltname,
                values=(entity),
                image=icon,
                open=open_init
            )
            _ELEMENT_DICT[iid] = elt

        if isinstance(elt,elements.Layout):
            make_layout_tree(elt, iid)

        for id, elt in screen.popupRegister.items():
            eltname = id
            elttype = str(elt.__class__.__name__)
            if elttype not in ELEMENT_ICONS_TK:
                    add_element_icon(elt)
            icon = ELEMENT_ICONS_TK.get(elttype, ELEMENT_ICONS_TK["default"])
            iid = elt.id
            entity = getattr(elt,"entity","None")

            if iid in treeview.get_children():
                treeview.delete(iid)

            treeview.insert(
                "",
                tk.END,
                iid = iid,
                text = eltname,
                values=(entity),
                image = icon,
                open = open_init
            )
            _ELEMENT_DICT[iid] = elt
            if isinstance(elt,elements.Layout):
                make_layout_tree(elt, iid)

    treeview.enable()
    return

def element_tree_selected(tree: Treeview, event, iid):
    "Figure out which element was selected in the element tree, and pass it on to the indicator function"
    if not EM_SETTINGS.getboolean(const.HIGHLIGHT_VAR_NAME):
        return
    
    eltList = []
    for iid in tree.selection():
        eltList.append(_ELEMENT_DICT[iid])
    
    highlight_element(*eltList)

def get_element(element_id) -> Optional[elements.Element]:
    return CORE.screen.elementRegister.get(element_id,None)

def tree_hover(tree, event, iid):
    if iid in CORE.screen.elementRegister:
        show_element_tip(tree, iid)

def tree_click(event: tk.Event):
    "Handles tree clicks not in an element"

    tree = event.widget
    if not tree.identify_row(event.y):
        for item in tree.selection():
            tree.selection_remove(item)
        highlight_element()

def tree_double_click(tree, event: tk.Event, iid):

    elt = get_element(iid)
    if elt:
        ElementWindow(elt)
    return


def tree_leave(event):
    tree = tree_frame.tree
    tree_frame.tree_tip.hide_tip()
    if not tree: return

    if tree_frame.last_hover:
        tree_frame.tree_tip.hide_tip()
        tree.item(tree_frame.last_hover, tags=[])
        tree_frame.last_hover = None

    return

def highlight_element(*element_list : "elements.Element"):
    "Pass items to draw a square around. Removes currently drawn squares first."
    _LOGGER.verbose(f"Removing {len(_INDICATOR_RECTANGLES)} rectangles.")
    for rect in canvas.find_above(const.SCREEN_TAG):
        ##If this throws errors after clearing: either use find_above and just delete them from there
        ##Just need to know the tag of the screen image itself.
        canvas.delete(rect)

    _INDICATOR_RECTANGLES.clear()

    if not EM_SETTINGS.getboolean(const.HIGHLIGHT_VAR_NAME):
        return

    for elt in element_list:
        _LOGGER.debug(f"Selected element in area {elt.area}")
        if elt.area == None or not elt.onScreen:
            continue
        [(x,y),(w,h)] = elt.area
        rect = canvas.create_rectangle(x,y,x+w,y+h, outline="red",
                                    dash=(5,2),width=5)
        _INDICATOR_RECTANGLES.append(rect)

def show_element_tip(tree: Treeview, event, element_id):
    
    if element_id not in CORE.screen.elementRegister:
        return

    elt = CORE.screen.elementRegister[element_id]

    tipText = (f"""{elt.__class__.__name__}: {element_id}\n
            Double click to view more info""")
    
    tree.tooltip.text = tipText
    
    tree.tooltip.show_tip()
    return

def highlight_setting(var_name, var_index, mode):

    highlight_element()

    val = window.hightlight_variable.get()

    tree_frame.update()

def __tree_list_setting(var_name, var_index, mode):
    if mode == 'write':
        window.trace_variable(var_name,var_index,mode)

    var_val = window.tree_list_variable.get()
    _LOGGER.debug(f"List has changed. Value is {var_val}")
    
    highlight_element() #Removes any highlights when the list changes

    if var_val == "Elements":
        tree_frame.entity_tree.pack_forget()
        tree_frame.element_tree.pack(anchor=tk.S, fill=tk.BOTH, expand=1)
        _LOGGER.debug("dont forget to rebuild the tree!")
    elif var_val == "Entities":
        tree_frame.element_tree.pack_forget()
        tree_frame.entity_tree.pack(anchor=tk.S, fill=tk.BOTH, expand=1)
        _LOGGER.debug("dont forget to rebuild the tree!")
    else:
        tree_frame.element_tree.pack_forget()
        tree_frame.entity_tree.pack_forget()
        if tree_frame.tree:
            tree_frame.tree_tip.hide_tip()
            tree_frame.unbind("<Enter>")
            tree_frame.unbind("<Leave>")
        return

    tree_frame.last_hover = False  

    tree_frame.tree_tip = ToolTip(tree_frame, bootstyle=const.TOOLTIP_STYLE)
    tree_frame.tree_tip.move_tip()
    tree_frame.tree_tip.hide_tip()



def import_funcs():

    element_tree = tree_frame.element_tree
    element_tree.on_select = element_tree_selected
    element_tree.on_double_click = tree_double_click
    element_tree.on_hover = show_element_tip

window.call_in_main_thread(import_funcs)


"Takes care of handling and setting the settings for the designer."

import logging
from pathlib import Path
from configparser import ConfigParser
from . import const 

logger = logging.getLogger(__name__)

EM_SETTINGS_FILE = Path(__file__).parent / "files" / "settings.ini"
"json file to save the emulator settings to."

DEFAULT_EM_SETTINGS = {
                    const.DARKMODE_VAR_NAME: False, 
                    const.SAVEAS_VAR_NAME: False, 
                    const.HIGHLIGHT_VAR_NAME: False, 
                    const.LIST_VAR_NAME: "None", 
                    "backlight": False, 
                    "battery_rnd": False, 
                    "network": False}
"The default settings to use when the settings file cannot be opened"

_section = "SETTINGS"
"Section for using EM_SETTING.get(SEC) (So you don't need to write SETTINGS each time.)"

EM_CONFIG = ConfigParser(DEFAULT_EM_SETTINGS, default_section=_section)
EM_CONFIG.read(EM_SETTINGS_FILE)

EM_SETTINGS = EM_CONFIG["SETTINGS"]
"Settings used for the emulator. Saved between sessions."

#Add a section per platform to keep settings for it

def save_settings():
    "Save the emulator settings to the settings file"
    logger.debug("Saving emulator settings")
    with open(EM_SETTINGS_FILE, 'w') as configfile:
        EM_CONFIG.write(configfile)


##HomeAssistantClient file
"""
This library manages elements from PSSM and linked entities from Home Assistant, and provides additional comound elements in the HAelements library.
Import it first, or at least before importing the element library to allow it to add to the PSSM default colors and fonts.
"""
from __future__ import annotations
from typing import Literal, TypedDict, TypeVar, Optional, Union
from pathlib import Path
import asyncio
from io import BytesIO
import logging

import requests
from PIL import Image

_LOGGER = logging.getLogger(__name__)

##Use this file to declare some more constants and extend a couple
class DomainError(ValueError):
    "The supplied entity is not of a valid domain."
    pass

##Mind the docstring when dealing with errors here (i.e. use from fromisoformat)

HAtimestrFormat = Literal["%Y-%m-%-dT%H:%M:%S.%Y+%H:%M"] #Literal[DEFAULT_HA_DT_FORMAT]
"Default Home Assistant datetime string typing"

EntityType = TypeVar("entity_id", bound=str)

stateDictType = TypedDict('stateDict', {"entity_id":str, "state": str, "attributes": dict, "last_changed": HAtimestrFormat,"last_reported": HAtimestrFormat, "last_updated": HAtimestrFormat, "context": dict})
"Typed dict for an entities state as passed from a trigger"

triggerDictType = TypedDict('triggerDictType',{"entity_id":str,"to_state": stateDictType, "from_state": Optional[stateDictType], "context":Optional[dict]})
"Typed dict for how triggers are passed"


actionCallDict = TypedDict('actionCallDict', {"id": int, "type": str, "domain": str, "service":str, "service_data": dict, "target": dict, "return_response": bool}, total=True)
"Typed dict for the message format of a HA action call via the websocket"

actionCallDict.__optional_keys__ = frozenset({"service_data", "target"})
actionCallDict.__required_keys__ = actionCallDict.__required_keys__.difference(actionCallDict.__optional_keys__)

serviceCallDict = actionCallDict

WeatherData = Literal["datetime", "cloud_coverage", "condition", "humidity", "temperature", "apparent_temperature", "dew_point", "pressure", "visibility", "wind_gust_speed", "wind_speed", "ozone", "uv_index", "wind_bearing"]
"Type hint with (most likely) all possible weather data in a weather entity's attributes."

async def request_image_threadsafe(image_url : str) -> tuple[Union[Image.Image, requests.Response],int]: #Union[tuple[Image.Image,Literal["status_code"]],tuple[requests.Response, Literal["status_code"]]]:    
    """
    Gets an image from a request.get response in a non-blocking manner.
    Method from: https://superfastpython.com/python-async-requests/

    Parameters
    ----------
    image_url : str
        url to get the image from

    Returns
    -------
    tuple[Image.Image | requests.response, status_code] | 
        If the status code is 200 (i.e. the request was succesfull) a tuple is returned with the gotten Image and the status code. 
        Otherwise a tuple with the full response and the status code is returned.
    """
    try:
        response = await asyncio.to_thread(requests.get, image_url)
    except (requests.exceptions.InvalidURL):
        _LOGGER.error(f"Cannot request image from {image_url}, invalid url")
        return (None, -1)

    if response.status_code == 200:

        ##Is this one threadsafe? Since it's not technically reading a file I'm not sure 
        ##Replaced it with a to_thread call to be sure
        # img = Image.open(BytesIO(response.content))
        respIO = BytesIO(response.content)
        img = await asyncio.to_thread(Image.open,respIO)
        return (img.copy(), response.status_code)
    else:
        _LOGGER.warning(f"Unable to get requested image")
        return (response, response.status_code)



import os
import logging
import socket
from typing import Optional
from pathlib import Path

from .fbink import API as FBInk

_LOGGER = logging.getLogger(__name__)

def is_wifi_connected() -> bool:
    return get_ip() != None

def get_ip() -> Optional[str]:
    """Gets the devices IP adress. Returns None if none found and sets the connected attribute appropriately"""
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except OSError as exce:
        _LOGGER.error(f"Error gettin IP: {exce}")
        IP = None
    finally:
        s.close()
    return IP

def get_SSID() -> Optional[str]:
    """Gets the name of the connected wifi network. Returns Wifi off if Wifi is off """
    
    if is_wifi_connected():
        network = (os.popen("iwgetid -r").read())
        return network
    else:
        return None

def get_mac() -> str:
    """Gets the devices mac adress"""
    ifconfig = (os.popen("ifconfig | grep eth0").read()).split()
    mac = ifconfig[-1]
    return mac

def kill_os():
    "Kills the os running on the ereader. Needs a reboot to get the os back."
    _LOGGER.info(f"Killing native {FBInk.platform} ui. Reboot the device the get it back.")
    sf = Path(__file__).parent / "scripts" / "kill-nickel.sh"
    os.system(str(sf))
    _LOGGER.info("OS killed")



"""Asyncio implementation of KIP
KIP : Kobo Input Python
Taken from
https://github.com/shermp/go-kobo-input
and translated very roughly to Python by Mavireck
Rewritten for asyncio by Slalamander
"""

import asyncio
import os,sys
import struct
from time import time
from fcntl import ioctl
import logging
from contextlib import suppress

from typing import *

from .fbink import API as fbink
from . import grabInput as grabber

class EventPacket(NamedTuple):
    time_sec: int
    time_u_sec: int
    event_type: int
    event_code: int
    event_value: int

DEFAULT_INPUT_DEVICE = "/dev/input/event1"
FORMAT = 'llHHI'
EVENT_SIZE = struct.calcsize(FORMAT)

VERBOSE = int(logging.DEBUG/2)

_LOGGER = logging.getLogger(__name__)

##From testing: seems that press = 0 and release = 0 gives the correct results for both pressure and touch?
##For Glo: button takes precedent. Pressure only to be utilised for more complicated stuff
##Also, (my) thumbs seem to mess up touch detection as they probably move too easily.
##Touchscreen is an IR one anyways, so the pressure event should not be pressure?
BUTTON_PRESS = 0
BUTTON_RELEASE = 1

PRESSURE_PRESS = 0
PRESSURE_RELEASE = 1

TOUCH_PRESSED = "PRESSED"
TOUCH_RELEASED = "RELEASED"

TOUCH_SHORT = "SHORT-TOUCH"
TOUCH_LONG = "LONG-TOUCH"

DEFAULT_HOLD_TIME = 0.5
DEFAULT_DEBOUNCE_TIME = 0.01

time_sec = TypeVar("TimeSec", bound=int)
time_u_sec = TypeVar("TimeUsec", bound=int)
ev_type = TypeVar("EvenType", bound=int)
ev_code = TypeVar("EventCode", bound=int)
ev_value = TypeVar("EventValue", bound=int)

event_types = {
	"Sync": 0,
	"Key": 1,
	"Absolute": 3,
}

event_key_codes = {
    "Btn2": 258,
	"BtnX": 307,
	"BtnY": 308,
	"Touch": 330,
}

event_abs_codes = {
    "X": 0,
    "Y": 1,
    "Hat0X": 16,
    "Hat0Y": 17,
    "Pressure": 24,
    "unknowns": [48,50,53,54,57]    
    ##Not sure what 50 and 57 do, values do not seem to change either
    ##Some of the other have to with pressure it seems, according to the GO code
}

evSyn = event_types["Sync"]
evAbs = event_types["Absolute"]
evKey = event_types["Key"]
btnTouch = event_key_codes["Touch"]

synReport            = 0
synDropped           = 3
synMTreport          = 2
absMTposX            = 53
absMTposY            = 54
absMTPressure        = 58
absMTtouchWidthMajor = 48

absX = {event_abs_codes["X"], event_abs_codes["Hat0X"], absMTposX}
absY = {event_abs_codes["Y"], event_abs_codes["Hat0Y"], absMTposY}
absPressure = {absMTPressure, absMTtouchWidthMajor, event_abs_codes["Pressure"]}

##The order in which a complete touch event seems to be transmitted:
##First the button press event is emitted
##Until the touch is moved, no coordinates come in (Hence it is not possible to differentiate between press and release, only the time a touch takes)
##X and Y are emitted
##Button release event is emitted

##X and Y do update when moving the finger, but that also emites the button release event. 
##Pressure does not seem to really correlate with this.
##I.e. moving your finger still emits pressure as false, and does not register a touch
##And when removing your finger, events are emitted again, so having a timeout waiting for new x/y updates does not really work either.


class InputQueue(asyncio.Queue):
    """Subclass of Asyncio Queue that passes on touch events.

    Asyncio implementation of KIP (Kobo Input Python). Simply await `InputQueue.get()` to get the touch events as they come in.
    Currently only supports press and release events, which can be evaluated using the `TOUCH_PRESSED` and `TOUCH_RELEASED` constants of this package.
    The first touch is never dispatched since it only emits a release event, no coordinates.

    General usage advice: debounce can be relatively big, as the screen is very sensitive. For holding, use a small finger, since it does not register those as moving as easily.
    I got the most consistent results using a key (I did not have a stylus on hand), though the smaller the object, the more likely it is to not cross any of the IR beams.

    Parameters
    ----------
    debounce_time : float, optional
        The minimum time between the press and release event for a touch to be valid, by default DEFAULT_DEBOUNCE_TIME
    long_click_time : float, optional
        The minimum time needed to hold a press for it to be dispatched as a long click, by default DEFAULT_HOLD_TIME
    input_device : str, optional
        Path the file that functions as the touch input, by default DEFAULT_INPUT_DEVICE
    loop : asyncio.AbstractEventLoop, optional
        the loop to attach to, by default None
    """    

    def __init__(self, debounce_time: float = DEFAULT_DEBOUNCE_TIME, long_click_time: float = DEFAULT_HOLD_TIME, input_device: str = DEFAULT_INPUT_DEVICE, 
                loop: asyncio.AbstractEventLoop = None):

        super().__init__(loop=loop)
        self._packet_queue = asyncio.Queue(loop=loop)
        self.__full_touch_event = asyncio.Event(loop=loop)
        self.__full_touch_event.set()
        if loop == None:
            loop = asyncio.get_running_loop()
        # self._output_queue = output_queue

        self._debounce_time = debounce_time
        self._long_click_time = long_click_time

        self._fileobject = open(input_device, "rb")   
        loop.add_reader(self._fileobject, self._read_inputdevice_line)
        self._read_task = loop.create_task(self._read_touch_packet())
        ioctl(self._fileobject, grabber.EVIOCGRAB(1), True)
        

    def __del__(self):
        self.release_input_grab()

    def release_input_grab(self):
        "Releases the input device"
        ioctl(self._fileobject, grabber.EVIOCGRAB(1), False)
        self._fileobject.close()
        print("Input device file closed")

    async def get(self) -> tuple[int,int,Union[TOUCH_PRESSED,TOUCH_RELEASED]]:
        return await super().get()

    def _read_inputdevice_line(self):
        line = self._fileobject.read(EVENT_SIZE)
        inp = struct.unpack(FORMAT, line)
        if not inp:
            _LOGGER.log(VERBOSE, f"binary read failed {inp}")
            print("binary read failed ", inp)
            return
        self._packet_queue.put_nowait(EventPacket(*inp))
        return
    
    async def _read_touch_packet(self):
        

        badPacket = False
        self.__eventdict = {}
        while not self._fileobject.closed:
            inp: EventPacket = await self._packet_queue.get()
            if inp.event_type == evSyn and inp.event_code == synDropped:
                badPacket = True
                continue
            if badPacket and inp.event_type == evSyn and inp.event_code == synReport:
                self.__eventdict = {}
                badPacket = False
                continue
            if not badPacket:
                self._decode_event(inp)
                await asyncio.sleep(0)
                if inp.event_type == evSyn and inp.event_code == synReport and not self.__full_touch_event.is_set():
                    asyncio.create_task(self._wait_for_event_dispatch())
        
    def _decode_packets(self, packets: list[EventPacket]):

        x, y = -1, -1
        touch = None
        pressure = None

        _LOGGER.log(5, f"Decoding {len(packets)} input event packets")
        for event in packets:
            if event.event_type == evKey: # and touch == None:
                #Some, but not all Kobo's report a BTN_TOUCH event
                #For the Glo HD I'm testing on, it doesn't even seem consistend
                if event.event_code == btnTouch and touch == None:
                    if event.event_value == BUTTON_PRESS:
                        touch = True
                    else:
                        touch = False
                    
                    print(f"Button touch set to {touch}")
            elif event.event_type == evAbs:
                if event.event_code in absX:
                    x = int(event.event_value)
                    print(f"X set to {x}")
                    
                elif event.event_code in absY:
                    y = int(event.event_value)
                    
                elif event.event_code in absPressure and pressure == None:
                    if event.event_value == PRESSURE_PRESS:
                        pressure = True
                    else:
                        pressure = False
                    print(f"Pressure set to {pressure}")

            # if x >= 0 and y >= 0 and touch != None and pressure != None:
            #     break
        
        if x < 0 or y < 0 or (touch == pressure == None):
            _LOGGER.log(VERBOSE, "Could not decode complete touch packet")
            return (x, y, TOUCH_RELEASED)

        if x >= 0 and y >= 0 and pressure != None:
            ##Kind of annoying, but there does not seem to be a release event when dragging.
            ##So probably pressure/sliding would need a seperate handler.
            pass

        ##Currently works, though it's rather finicky with slight movements causing a release

        if touch != None:
            touch_val = TOUCH_PRESSED if touch else TOUCH_RELEASED
        else:
            touch_val = None

        return (x, y, touch_val)

    def _decode_event(self, event: EventPacket):

        if event.event_type == evKey:
            #Some, but not all Kobo's report a BTN_TOUCH event
            #For the Glo HD I'm testing on, it doesn't even seem consistent
            if event.event_code == btnTouch:
                if event.event_value == BUTTON_PRESS:
                    self.__eventdict["touch-true"] = True
                    self.__full_touch_event.clear()
                else:
                    self.__eventdict["touch-false"] = False
                    self.__full_touch_event.set()
                
        elif event.event_type == evAbs and "touch-true" in self.__eventdict:
            if event.event_code in absX:
                x = int(event.event_value)
                self.__eventdict["x"] = x
            elif event.event_code in absY:
                y = int(event.event_value)
                self.__eventdict["y"] = y
            elif event.event_code in absPressure:
                if event.event_value == PRESSURE_PRESS:
                    pressure = True
                else:
                    pressure = False
                self.__eventdict["pressure"] = pressure

    def _is_complete_event(self) -> bool:
        return not bool({"x","y","touch-true", "touch-false"} - set(self.__eventdict.keys()))

    def _rotate_coordinates(self, x: int, y: int):
        "Rotates the x and y coordinates received such that the upper left corner registers as (0,0) and x is the horizontal axis and y is the vertical axis"

        ##canonical 0 is upright portrait, which will be considered the general starting point BUT device has 0 rota
        rota = fbink.current_rota
        rx = x
        ry = y
        if rota == 0:
            ##No need to transpose them
            rx = x
            ry = y
        elif rota == 1:
            rx = y
            ry = fbink.screen_height - x
        elif rota == 2:
            rx = fbink.screen_width - x
            ry = fbink.screen_height - y
        elif rota == 3:
            rx = fbink.screen_width - y
            ry = x

        _LOGGER.log(VERBOSE, f"Rotation of {rota} (Canonical: {fbink.current_rota_canonical}), Original (x,y): {(x,y)}, rotated to (rx,ry): {(rx, ry)}.")
        return (rx, ry)

    async def _wait_for_event_dispatch(self):
        ##When gathering a touch, waits for the touch release event
        ##Does not dispatch if it completes within debounce
        ##Then determines if it is a long click or a short click
        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(self.__full_touch_event.wait(),
                                timeout=self._debounce_time)
            _LOGGER.log(VERBOSE, "Debounced touch")
            return
        
        await asyncio.sleep(0)

        try:
            await asyncio.wait_for(self.__full_touch_event.wait(),
                    timeout=self._long_click_time)
        except asyncio.TimeoutError:
            t = TOUCH_LONG
            await self.__full_touch_event.wait()
        else:
            t = TOUCH_SHORT

        if not self._is_complete_event():
            return

        x = self.__eventdict["x"]
        y = self.__eventdict["y"]
        self.__eventdict = {}

        _LOGGER.log(VERBOSE, f"Passing touch at {(x,y)} as {t}")
        x,y = self._rotate_coordinates(x,y)
        await self.put((x,y,t))





"""
I have not made any of this stuff, I only used resources from Whizse's Github:
https://github.com/whizse/exclusive-keyboard-access
And I only kept the necessary
"""

import ctypes
from time import sleep
import fcntl


_IOC_NRBITS = 8
_IOC_TYPEBITS = 8
_IOC_SIZEBITS = 14
_IOC_DIRBITS = 2

_IOC_NRMASK = (1 << _IOC_NRBITS) - 1
_IOC_TYPEMASK = (1 << _IOC_TYPEBITS) - 1
_IOC_SIZEMASK = (1 << _IOC_SIZEBITS) - 1
_IOC_DIRMASK = (1 << _IOC_DIRBITS) - 1

_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS

IOC_NONE = 0
IOC_WRITE = 1
IOC_READ = 2

def IOW(type, nr, size):
    return IOC(IOC_WRITE, type, nr, IOC_TYPECHECK(size))

def IOC_TYPECHECK(t):
    result = ctypes.sizeof(t)
    assert result <= _IOC_SIZEMASK, result
    return result

def IOC(dir, type, nr, size):
    assert dir <= _IOC_DIRMASK, dir
    assert type <= _IOC_TYPEMASK, type
    assert nr <= _IOC_NRMASK, nr
    assert size <= _IOC_SIZEMASK, size
    return (dir << _IOC_DIRSHIFT) | (type << _IOC_TYPESHIFT) | (nr << _IOC_NRSHIFT) | (size << _IOC_SIZESHIFT)


EVIOCGRAB = lambda len: IOW(ord('E'), 0x90, ctypes.c_int)

def grab():
    fcntl.ioctl(fd, EVIOCGRAB(1), True)

def ungrab():
    fcntl.ioctl(fd, EVIOCGRAB(1), False)


"Constants for HA client designer"

from typing import Literal
from mdi_pil import mdiType

__entity_icons : dict[Literal["domain"],mdiType] = {
    "default": "mdi:puzzle",
    
    ##All entities found in https://developers.home-assistant.io/docs/core/entity
    "alarm_control_panel": "mdi:shield-lock",
    "binary_sensor": "mdi:radiobox-indeterminate-variant",
    "button": "mdi:button-pointer",
    "calendar": "mdi:calendar-blank",
    "camera": "mdi:camera",
    "climate": "mdi:radiator-disabled",
    "conversation": "mdi:chat",
    "cover": "mdi:blinds",
    "date": "mdi:calendar",
    "datetime": "mdi:calendar-clock",
    "device_tracker": "mdi:map-marker-account",
    "event": "mdi:star-four-points-circle",
    "fan": "mdi:fan",
    "humidifier": "mdi:air-humidifier",
    "image": "mdi:image",
    "lawn_mower": "mdi:robot-mower",
    "light": "mdi:lightbulb",
    "lock": "mdi:lock",
    "media_player": "mdi:multimedia",
    "notify": "mdi:message",
    "number": "mdi:numeric",
    "remote": "mdi:remote",
    "scene": "mdi:palette",
    "select": "mdi:form-select",
    "sensor": "mdi:speedometer",
    "siren": "mdi:alarm-light",
    "stt": "mdi:account-voice",
    "switch": "mdi:toggle-switch",
    "text": "mdi:form-textbox",
    "time": "mdi:clock",
    "todo": "mdi:format-list-checks",
    "tts": "mdi:microphone-message",
    "update": "mdi:update",
    "vacuum": "mdi:robot-vacuum",
    "valve": "mdi:valve",
    "wake_word": "mdi:microphone-question",
    "water_heater": "mdi:kettle",
    "weather": "mdi:cloud",
    "zone": "mdi:map-marker",

    ##Some domains provided by core that are not in the developer list (may be inconclusive)
    "automation": "mdi:robot",
    "person": "mdi:account",
    "sun": "mdi:weather-sunny",
    "plant": "mdi:flower",
    "script": "mdi:script",

    ##Unique Helper domains
    "counter": "mdi:counter",
    "timer": "mdi:timer",
    "schedule": "mdi:calendar-star"

}

__helper_map = {"input_button": "button", "input_boolean": "switch", "input_number": "number",
                "input_select": "select", "input_text": "text", "input_datetime": "datetime"}

for helper, dom in __helper_map.items():
    __entity_icons[helper] = __entity_icons[dom]

ENTITY_ICONS = __entity_icons
"Icons for entity domains in the treeview"